'use strict'

/*
http://frutbunn.com  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.

A very simple keyboard class for flatform games etc.
*/

// keymap = an object listing keys to check for and a mnemonic to associate it with, for
// example: {"65": "MoveLeft", "66": "MoveLeft, "67": "MoveRight", "68": "MoveRight"}
// 
// allowDefault is optional, if true will allow default browser key press events to also trigger.
//
// To check if a key in a mnemonic group has been pressed:
// var o = new FruKeyboard( {"65": "MoveLeft", "66": "MoveLeft, "67": "MoveRight", "68": "MoveRight"} );
// var aMoveLeftKeyIsDown = o.isDown("MoveLeft"); // will return true or false if key 65 (A) or 66 (B)
// is current down.
function FruKeyboard(keyMap, allowDefault) {
	var that = this;

	var state = {
		down: null,			// Array that holds 'keydown' state of any key
		map: null,			// Object that holds a mapping between keys and mnemonics
		mnemonic: null,		// Object that holds state of mnemonics

		allowDefault: null,

		aKeyIsDown: false
	}

	function makeMap(keyMap) {
		state.map = {};
		
		for(var n in keyMap)
			state.map[n] = keyMap[n];
	}

	function makeMnemonic(keyMap) {
		state.mnemonic = {};

		for(var n in keyMap)
			state.mnemonic[ keyMap[n] ] = false;
	}

	function constructor() {
		state.allowDefault = allowDefault ? true : false;
		state.down = [];
		makeMap(keyMap);
		makeMnemonic(keyMap);

		window.addEventListener('keydown', keyPressEvent, false);
		window.addEventListener('keyup', keyPressEvent, false);
	}

	function keyPressEvent(event) {
		if (event.type == 'keydown')
			state.aKeyIsDown = true;
		else
			state.aKeyIsDown = false;

		if (event.type == 'keydown' || event.type == 'keyup') {
			var b = (event.type == 'keydown');

			state.down[event.keyCode] = b;
			state.mnemonic[ state.map[event.keyCode] ] = b;
		}

		//console.log(event.keyCode);

		if (! state.allowDefault)
			event.preventDefault();

	}

	constructor();

	return {
		get aKeyIsDown() { return state.aKeyIsDown ? true : false; },
		isDown: function(name) { return state.mnemonic[name] ? true : false; },
		trigger: keyPressEvent
	}

}