'use strict'

/*
FruTools version 0.03

http://frutbunn.com  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

var FruTools = function() { 

	function Coordinate(x, y) {
		this.x = x || 0;
		this.y = y || 0;
	}


	function Bob(x, y, w, h) {
		this.x = x || 0;
		this.y = y || 0;
		this.w = w || 0;
		this.h = h || 0;
	}


	function Tile(canvasWidth, tileWidth, tileHeight, horizontalTileCount) {
		var state = {
			horizontalTileCount: null,

			tileSize: {
				width: null,
				height: null
			}
		};

		function constructor() {
			if (horizontalTileCount)
				state.horizontalTileCount = horizontalTileCount;
			else
				state.horizontalTileCount = Math.floor(canvasWidth/tileWidth);

			state.tileSize.height = Number(tileHeight);
			state.tileSize.width = Number(tileWidth);
		}

		function calcTilePosition(index, horizontalTileCount) {
			var r = new Coordinate(Math.floor(index%horizontalTileCount), Math.floor(index/horizontalTileCount) );
			
			return r;
		}

		function position(index) {
			var b = calcTilePosition(index, state.horizontalTileCount);

			b.x*= state.tileSize.width;
			b.y*= state.tileSize.height;

			return new Bob(b.x, b.y, state.tileSize.width, state.tileSize.height);
		}

		constructor();

		return {
			get width() { return state.tileSize.width; },
			get height() { return state.tileSize.height; },

			position: position
		}
	}


	function Animation(duration, countOrIndices, loop) {
		var state = {
			count: null,
			duration: null,
			loop: null,
			indices: null
		};

		function constructor() {
			if (countOrIndices instanceof Array) {
				state.indices = [].concat(countOrIndices);
				state.count = state.indices.length;
			} else {
				state.indices = null;
				state.count = countOrIndices;
			}

			state.duration = duration;
			state.loop = loop ? true : false;
		}

		function currentFrame(currentTimeInSeconds) {
			// current time - Fixed to duration boundry, and always positive number
			var ct = Math.max(0., currentTimeInSeconds);
			
			if (!state.loop && ct>=state.duration)
				if (state.indices) 
					return state.indices[state.count-1];
				else
					return state.count-1;

			else
				ct %= state.duration;

			// current percentage - if duration is zero, fix to zero
			var cp = state.duration == 0 ? 0 : Math.min(ct/state.duration, 1);

			// current frame
			var cf = Math.floor((state.count)*cp);

			// If indices supplied, return that, otherwise return frame number
			if (state.indices)
				return state.indices[cf];
			else
				return cf;
		}


		constructor();

		return {
			current: currentFrame,
			get count() {return state.count; },
			get duration() {return state.duration; }
		};

	}


	// API:
	
	return {
		Coordinate: Coordinate,
		Bob: Bob,

		Tile: Tile,
		Animation: Animation
	}
}();



var fruCookie = {
            set: function(name, value, path, days) {
                days = typeof days === 'undefined' ? ""
                 : "expires=" + new Date(+new Date() + 24*60*60*days*1000 ).toUTCString() + "; ";
                path = typeof path === 'undefined' ? "" : "path=" + path + "; ";

                document.cookie = name + "=" + JSON.stringify(value) + "; " + days + path;
            },

            get: function(name) {
                var c = String(document.cookie);
                var a = c == "" ? null : c.split(';');

                for(var i in a) {
                    var p = String(a[i]).trim().match(/^(.*?)=(.*)/, '$1, $2');
                    if (! p) continue;

                    var key = p[1];
                    if (name == key) {
                        var value = p[2];
                        try{ value = JSON.parse(value); } catch(e) {}

                        return value;
                    }
                }

                return null;
            }

}
