'use strict'

/*
http://frutbunn.com  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

var cuddlyMenuScreen = (function() {

	var AUTO_PILOT_ACTIVATE_DURATION = 60*60*2;

	var player = new music("YM");

	var state = {
		keyboard: null,
		globalTime: new FruTimer(),
		carebaresGlobalTime: new FruTimer(),

		stepping: null,

		cuddlyLevel: null,
		scrollerLevel: null,

		// Everything in model will be subject to linear interpolation by fru_time_stepping.js
		model: {
			position: {	// Position of dude on the scrolling map
				x: 320,
				y: 450
			},

			direction: 1, // Direction red dude is facing 0 = left, 1 = right
			moving: 0,
			thrusting: 0,

			justLanded: 0,
			bounceDisplacement: 0,

			thrustSpeed: 0,
			fallingSpeed: 0,

			currentFrame: 6,

			carebears: {
				x: [0,0,0,0,0,0,0,0,0,0,0,0],
				y: [0,0,0,0,0,0,0,0,0,0,0,0]
			},

			scrollerPosition: 0
		},

		// These are updated in the model, but do not need to be subject to linear interpolation
		autoPilot: {
			//wasActiveLastTime: false,
			activateIn: AUTO_PILOT_ACTIVATE_DURATION,

			dontThrustForAwhile: 0,
			timeSinceLastThrust: 200,
			waitToLoad: 80,
			nowLoadScreen: false,
			stuck: false,
			keepWalkingForAwhile: 0,

			nextScreen: 0
		},


		thrustOff: 0 // Hack to get red dude to bounce on top of screen - just noticed original does this...
	};


	var resource = {
		totalImagesToLoad: 4,

		image: null,	// image is an object setup in bootstrap()

		ctx: {
			game: null,
			background: null
		},

		tile: {
			dude: new FruTools.Tile(640, 64, 64),
			carebears: new FruTools.Tile(384, 32, 20)
		},

		animation: {
			dude: {
				move: {
					right: new FruTools.Animation(.35, [2, 3, 4, 5, 6, 7, 8, 9], true),
					left: new FruTools.Animation(.35, [12, 13, 14, 15, 16, 17, 18, 19], true)
				},
				thrust: {
					right: new FruTools.Animation(.075, [0, 1], true),
					left: new FruTools.Animation(.075, [10, 11], true)
				}
			}
		},

		background: {
			tile: new FruTools.Tile(320, 16*2, 16*2, 10)
		},

		bouncing: {
				animation: [0,3,5,6,5,3,0,1,2,3,2,1,0],
				speed: 7
		},
/*
		demoScreens: [
			new DemoScreenLocation(27, 7, "BIG_SPRITE"),
			new DemoScreenLocation(63, 16, "COLORSHOCK_II"),
			new DemoScreenLocation(87, 6, "NO_NAME_1"),
			new DemoScreenLocation(128, 1, "MEGA_SCROLLER"),
			new DemoScreenLocation(144, 8, "SPREADPOINT"),
			new DemoScreenLocation(168, 16, "DIGI_DEMO"),
			new DemoScreenLocation(197, 2, "LED_SCROLLER"),
			new DemoScreenLocation(294, 2, "DOC"),
			new DemoScreenLocation(351, 13, "FULLSCREEN"),
			new DemoScreenLocation(373, 5, "STARWARS_DEMO"),
			new DemoScreenLocation(376, 16, "KNUCKLE_BUSTER"),
			new DemoScreenLocation(416, 14, "DNA_DEMO"),
			new DemoScreenLocation(438, 1, "NO_NAME_2")
		]
*/

		demoScreens: [
			new DemoScreenLocation(27, 7, "BIG_SPRITE"),
			new DemoScreenLocation(63, 16, "COLORSHOCK_II"),
			new DemoScreenLocation(87, 6, "NO_NAME_1"),
			new DemoScreenLocation(128, 1, "MEGA_SCROLLER"),
			new DemoScreenLocation(168, 16, "DIGI_DEMO"),
			new DemoScreenLocation(144, 8, "SPREADPOINT"),
			new DemoScreenLocation(197, 2, "LED_SCROLLER"),
			new DemoScreenLocation(294, 2, "DOC"),
			new DemoScreenLocation(351, 13, "FULLSCREEN"),
			new DemoScreenLocation(373, 5, "STARWARS_DEMO"),
			new DemoScreenLocation(376, 16, "KNUCKLE_BUSTER"),
			new DemoScreenLocation(416, 14, "DNA_DEMO"),
			new DemoScreenLocation(438, 1, "NO_NAME_2"),

			new DemoScreenLocation(416, 14, "DNA_DEMO"),
			new DemoScreenLocation(376, 16, "KNUCKLE_BUSTER"),			
			new DemoScreenLocation(373, 5, "STARWARS_DEMO"),
			new DemoScreenLocation(351, 13, "FULLSCREEN"),
			new DemoScreenLocation(294, 2, "DOC"),
			new DemoScreenLocation(197, 2, "LED_SCROLLER"),
			new DemoScreenLocation(144, 8, "SPREADPOINT"),
			new DemoScreenLocation(168, 16, "DIGI_DEMO"),
			new DemoScreenLocation(128, 1, "MEGA_SCROLLER"),
			new DemoScreenLocation(87, 6, "NO_NAME_1"),
			new DemoScreenLocation(63, 16, "COLORSHOCK_II"),

			new DemoScreenLocation(171, 16, "CREDITS")
		]

	}

	function anImageLoaded() {
	    if (--resource.totalImagesToLoad>0)
	        console.log("An image loaded. (Still waiting for " + resource.totalImagesToLoad + " resources to load).");
	    else
			console.log("All resources loaded...");
	}

	function loadImage(url) {
		console.log("Loading image: " + url);
		var img = document.createElement('img');
		img.src = url;
		img.addEventListener('load', anImageLoaded);

		return img;
	}

	function createCanvasElement(width, height) {
		var canvas = document.createElement('canvas');
		canvas.width  = width;
		canvas.height = height;
		return canvas.getContext("2d");
	}

	var requestAnimFrame = (function() {
	   	return window.requestAnimationFrame || window.webkitRequestAnimationFrame || window.mozRequestAnimationFrame 
	   	 || window.oRequestAnimationFrame || window.msRequestAnimationFrame
	   	 || function(callback, element) { window.setTimeout(callback, 1000/60); };
	})();

	// The first function that runs - see index.html
	function bootstrap() {

		// Source canvas
		mycanvas = new canvas(768, 536);
		cuddlyDemosShared.initCrtEmulator(mycanvas, "main");

        //	-	-	-

		resource.ctx.game = createCanvasElement(320*2, 144*2);
		resource.ctx.background = createCanvasElement(320*2 + 16*2, cuddlyMap.length*16*2 + 16*2); // Background width + 1 tile

		//	-	-	-

		// Do not move this from bootstrap(). Otherwise the onload event in the HTML will not
		// be triggered until these images have loaded
		resource.image = {
			carebears: loadImage(basepath + 'carebears.png'),
			dude: loadImage(basepath + 'dude.png'),
			tiles: loadImage(basepath + 'tiles.png'),
			chrome: loadImage(basepath + 'chrome.png')
		}

		//	-	-	-

		// Initialize Karl's loader screen
		loaderInit();

		//	-	-	-

		// Show a black screen until all images have been loaded
		goLoadingImages();
	}

	// Black screen while loading image resources
	function goLoadingImages() {
		if (state.globalTime.seconds.current>.1 && resource.totalImagesToLoad<=0) {
			main();

			return;
		}

		//	-	-	-

		mycanvas.fill('#000000');

		//	-	-	-

		crtEmulator.updateFrame();
		requestAnimFrame(goLoadingImages);
	}

	function main() {
		state.keyboard = new FruKeyboard({
			"37": "left",
			"90": "left",

			"39": "right",
			"88": "right",
			
			"38": "thrust",
			"13": "thrust",
			
			"32": "load",

			"98": "down",
			"104": "up",

			"82": "reset"
		});

		state.stepping = new FruTimeStepping(1/60, state.model, integrate, render, state);
		state.cuddlyLevel = new FruRenderMap(resource.image.tiles, cuddlyMap, 16*2
		 , resource.ctx.game.canvas,    0, 0, 320*2, 144*2);
		
		var scrollerMap = buildScrolltextData(scrolltextData); // scrolltextData is in data.js
		state.scrollerLevel = new FruRenderMap(resource.image.chrome, scrollerMap
		 , [16*2, 40*2], mycanvas.canvas, 64, 400, 640, 80);
		
		// Precalculate background scrolling area
		fillBackground();

		// See if coming back to a previous position on the map / scrolltext
		try {
			var cookieData = fruCookie.get("cuddlyMenu");
			if (cookieData != null) {
				state.model.position.x = cookieData.locationX;
				state.model.position.y = cookieData.locationY-50;
				state.model.scrollerPosition = cookieData.scrollerPosition;
			}
		} catch(e) {}

		getAutoPilotCookie();

		//	-	-	-

		state.globalTime.reset();
		state.carebaresGlobalTime.reset();

		sineSprites.init(state.carebaresGlobalTime, resource.image.carebears, resource.tile.carebears, resource.ctx.game);

		// -	-	-

		// Run Karl's loader screen in loader.js
		loader("MENU");
	}

	// Karl's code will call this function to run the menu
	function runMenu() {

//		setTimeout(function(){
			player.LoadAndRun(basepath + 'menu.ym');
			go();
//		}, 750);

	}

	function buildScrolltextData(text) {
		var scrollerCharWidth = [
			3, 1, 2, 3, 3, 3, 3, 1, 1, 1, 3, 2, 1, 2, 1, 2, 3, 2, 2, 2, 2, 2, 
			2, 2, 2, 2, 1, 1, 3, 2, 3, 2, 3, 
			3, 2, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 3, 2, 3, 2, 3, 2, 2, 2, 2, 3, 3, 3, 2, 2,
			3, 3, 3, 3, 3, 3
		];

		var scroller = [];
		for(var i=0; i<text.length; i++) {
			var char = text.charCodeAt(i);
			var p = (char-32);
			var blocks = scrollerCharWidth[p];

			for(var j=0; j<blocks; j++)
				scroller.push(p*3 + j);
		}

		var map = [];
		map.push(scroller);

		return map;
	}

	//	-	-	-

	function DemoScreenLocation(x, y, screen) {
		this.x = x;
		this.y = y;
		this.screen = screen;
	}

	function loadScreen() {
		var load = state.keyboard.isDown("load");

		if (state.autoPilot.nowLoadScreen==true) {
			setAutoPilotCookie(state.autoPilot.nowLoadScreen);
			load=true;
		}

		if (load) {
			var p = new FruTools.Coordinate(
			 Math.floor(state.model.position.x/32), Math.floor(state.model.position.y/32)
			);

			for(var i in resource.demoScreens) {
				var t = resource.demoScreens[i];

				if (p.x>=t.x && p.x<=t.x+3 && p.y>=t.y && p.y<t.y+3) {

					// Store position of dude in a cookie so we know where we were when we come back
					try {
						fruCookie.set("cuddlyMenu", {
							locationX: state.model.position.x,
							locationY: state.model.position.y,
							scrollerPosition: state.model.scrollerPosition
						}, "/");
					} catch(e) {}

					// Stop music
					player.LoadAndRun(basepath + 'mute.ym');

					// Clear screen
					mycanvas.fill('#000000');
					crtEmulator.updateFrame();

					// Run Karl's loader screen
					setTimeout(function() {
						loader(t.screen);
					}, 500);

					return true;
				}

			}

		}

		return false;
	}

	//	-	-	-

	// Main game loop
	function go() {
		var time = state.globalTime.seconds.current;
		state.stepping.go(time);

		if (state.keyboard.isDown("reset")) {
			window.open('cuddly_reset.html' , '_self');
		} else if (! loadScreen() ) {
			crtEmulator.updateFrame();
			requestAnimFrame(go);
		}

	}

	//	-	-	-

	function getAutoPilotCookie() {
		var wasActiveLastTime = false;

		try {
			var cookieData = fruCookie.get("cuddlyMenu-autoPilot");
			if (cookieData != null) {
				wasActiveLastTime = cookieData.active;
				state.autoPilot.nextScreen = cookieData.nextScreen;
			}
		} catch(e) {}

		if (wasActiveLastTime)
			state.autoPilot.activateIn = 120;
	}

	function setAutoPilotCookie(active) {
		var nextScreen = state.autoPilot.nextScreen+1;
		if (nextScreen>=resource.demoScreens.length-1) // -1 so we don't load the credit screen on auto pilot
			nextScreen=0;

		try {
			fruCookie.set("cuddlyMenu-autoPilot", {
				active: active,
				nextScreen: nextScreen
			}, "/");
		} catch(e) {}
	}

	function autoPilot(model) {
		var left = false;
		var right = false;
		var thrust = false;

		var screen = state.autoPilot.nextScreen;

		var target = new FruTools.Coordinate(
			resource.demoScreens[screen].x*32 + 64,
			resource.demoScreens[screen].y*32 + 64
		);

		var current = new FruTools.Coordinate(
			model.position.x,
			model.position.y
		);

		current.x &= ~0xf;
		target.x &= ~0xf;

		var onALevel = haveLanded(model);

		if (onALevel && (current.y&~0x1f)<(target.y&~0xf) && current.y<550) {
			state.autoPilot.stuck = true;
			state.autoPilot.keepWalkingForAwhile = 30;
		}

		if (current.y>=550)
			state.autoPilot.stuck = false;

		if (state.autoPilot.stuck) {
			if (model.direction==0)
				left = true;
			else
				right = true;

			if (!onALevel) {

				state.autoPilot.keepWalkingForAwhile--;

				if (state.autoPilot.keepWalkingForAwhile<=0)
					state.autoPilot.stuck = false;
			}
		} else {
			if (current.x==target.x) {
				state.autoPilot.waitToLoad--;
			}

			if (state.autoPilot.waitToLoad<=0) {
				state.autoPilot.nowLoadScreen = true;
			}

			if ( (current.x)<(target.x) )
				right=true;

			if ( (current.x)>(target.x) )
				left=true;

			if ( (current.y > target.y && state.autoPilot.dontThrustForAwhile<70) )
				thrust=true;
		}


		if (state.autoPilot.timeSinceLastThrust>0)
			state.autoPilot.timeSinceLastThrust--;

		if (state.autoPilot.timeSinceLastThrust<=0)
			state.autoPilot.timeSinceLastThrust=Math.floor(Math.random() * 100) + 100; //200;

		if (state.autoPilot.dontThrustForAwhile>0)
			state.autoPilot.dontThrustForAwhile--;

		if (state.autoPilot.dontThrustForAwhile<=0)
			state.autoPilot.dontThrustForAwhile = 100;
		
		return {
			left: left,
			right: right,
			thrust: thrust
		};

	}

	function integrate(model, t, dt, userObject) {
		var left = state.keyboard.isDown("left");
		var right = state.keyboard.isDown("right");
		var thrust = state.keyboard.isDown("thrust");

		//	-	-	-

		if (state.keyboard.aKeyIsDown) {
			if (state.autoPilot.activateIn<=0) {
				setAutoPilotCookie(false);
			}

			state.autoPilot.activateIn = AUTO_PILOT_ACTIVATE_DURATION;
		} else {
			state.autoPilot.activateIn--;
		}


		if (state.autoPilot.activateIn<=0) {
			var autoMovement = autoPilot(model);
			left = autoMovement.left;
			right = autoMovement.right;
			thrust = autoMovement.thrust;			
		}

		//	-	-	-

		if (model.bounceDisplacement>0)
			model.bounceDisplacement-=1;

		if (left && !right) {
			model.position.x -= resource.bouncing.speed;
			model.direction = 0;
			model.moving = 1;
		}

		if (right && !left) {
			model.position.x += resource.bouncing.speed;
			model.direction = 1;
			model.moving = 1;
		}

		if (!right && !left) {
			model.moving = 0;
		}

		if (thrust) {
			model.thrusting = 1;

			if (model.thrustSpeed<=0)
				model.thrustSpeed = 3;

			if (model.thrustSpeed<8 )
				model.thrustSpeed += 1;

			model.fallingSpeed = 0;
		} else {
			model.thrusting = 0;

			if (model.position.y<32)
				model.thrustSpeed = 0;
		}


		if (model.thrustSpeed<=0 && model.justLanded!=1) { 
			if (model.fallingSpeed<resource.bouncing.speed)
				model.fallingSpeed +=  .5;
		}


		for(var i=0; i< Math.floor(model.fallingSpeed)+1; i++) {
			if (! haveLanded(model) ) {
				model.position.y += 1; 
				model.justLanded = 0;
			} else {
				if (model.justLanded < 1) {
					model.justLanded = 1;

					model.bounceDisplacement = resource.bouncing.animation.length;
					model.thrustSpeed = 0;
				}

			}

		}

		if (model.thrustSpeed>0) 
			model.thrustSpeed -= .5;

		if (model.thrustSpeed>0 & !state.thrustOff)
			model.position.y -= Math.floor(model.thrustSpeed);

		// clipping
		var DUDE_WIDTH = resource.tile.dude.width;
		var DUDE_HEIGHT = resource.tile.dude.height;
		var MAP_WIDTH_IN_PIXELS = state.cuddlyLevel.widthInPixels;
		var MAP_HEIGHT_IN_PIXELS = state.cuddlyLevel.heightInPixels - (32-4);

		// Same as Cuddly Menu
		if (model.position.x<9*32) 
			model.position.x = 9*32;
		
		if (model.position.y<0) {
			model.position.y = 0;
			state.thrustOff = 10;
		}
		if (state.thrustOff>0) {
			state.thrustOff--;
			model.position.y += 2; // Falling starts quite slow, so counteract that when bouncing on the top of the screen
		}
		
		if (model.position.x>=455*32) 
			model.position.x = 455*32;
		
		if (model.position.y>=MAP_HEIGHT_IN_PIXELS-DUDE_HEIGHT) 
			model.position.y = MAP_HEIGHT_IN_PIXELS-DUDE_HEIGHT;


		// Move scrolly message at bottom
		integrateScroller(model, t, dt, userObject);
	}

	function integrateScroller(model, t, dt, userObject) {
		model.scrollerPosition += 8;
	}

	function haveLanded(model) {
		var FLOOR_TILE = 69;

		var dudePosition = new FruTools.Coordinate(model.position.x, model.position.y);
		var topLeftBlock = new FruTools.Coordinate( Math.floor(dudePosition.x/32), Math.floor(dudePosition.y/32) );
		
		var collision = false;

		if ( (cuddlyMap[topLeftBlock.y+2][topLeftBlock.x]==FLOOR_TILE
		 || cuddlyMap[topLeftBlock.y+2][topLeftBlock.x+1]==FLOOR_TILE
		 || ( topLeftBlock.x<cuddlyMap[0].length-2 && cuddlyMap[topLeftBlock.y+2][topLeftBlock.x+2]==FLOOR_TILE )
		 ) && model.position.y%32 == 0) 
			collision = true;

		return collision;
	}

	function render(model, t, dt, userObject) {
		mycanvas.fill('#000000');

		var DUDE_WIDTH = resource.tile.dude.width;
		var DUDE_HEIGHT = resource.tile.dude.height;
		var MAP_WIDTH_IN_PIXELS = state.cuddlyLevel.widthInPixels;
		var MAP_HEIGHT_IN_PIXELS = state.cuddlyLevel.heightInPixels - (32-4);

		var canvas = resource.ctx.game.canvas;

		var dudePosition = new FruTools.Coordinate(Math.round(model.position.x), Math.round(model.position.y) );
		var dudeScreenPosition = new FruTools.Coordinate();
		var mapPosition = new FruTools.Coordinate();

		// Calculate position of dude, map and background
		if (dudePosition.x <= canvas.width/2 - DUDE_WIDTH/2) {
			dudeScreenPosition.x = dudePosition.x;
			mapPosition.x = 0;
		} else if (dudePosition.x > MAP_WIDTH_IN_PIXELS - canvas.width/2 - DUDE_WIDTH/2) {
			dudeScreenPosition.x = canvas.width - (MAP_WIDTH_IN_PIXELS-dudePosition.x);
			mapPosition.x = MAP_WIDTH_IN_PIXELS - canvas.width;
		} else {
			dudeScreenPosition.x = canvas.width/2 - DUDE_WIDTH/2;
			mapPosition.x = dudePosition.x - (canvas.width/2 - DUDE_WIDTH/2);
		}

		if (dudePosition.y <= canvas.height/2 - DUDE_HEIGHT/2) {
			dudeScreenPosition.y = dudePosition.y;
			mapPosition.y = 0;
		} else if (dudePosition.y > MAP_HEIGHT_IN_PIXELS - canvas.height/2 - DUDE_HEIGHT/2) {
			dudeScreenPosition.y = canvas.height - (MAP_HEIGHT_IN_PIXELS-dudePosition.y);
			mapPosition.y = MAP_HEIGHT_IN_PIXELS - canvas.height;
		} else {
			dudeScreenPosition.y = canvas.height/2 - DUDE_HEIGHT/2;
			mapPosition.y = dudePosition.y - (canvas.height/2 - DUDE_HEIGHT/2);
		}

		// Calculate dude's landing bounce
		var bounce=resource.bouncing.animation[resource.bouncing.animation.length-Math.floor(model.bounceDisplacement)-1];
		bounce *= ( (model.fallingSpeed<3 ? 3 : model.fallingSpeed) / resource.bouncing.speed) * .9;

		// Calculate red dude frame of animation
		var dudeFrame = state.model.currentFrame;
		if (Math.ceil(model.thrusting) ) {
			if (Math.ceil(model.direction)<1)
				dudeFrame = resource.animation.dude.thrust.left.current(t);
			else
				dudeFrame = resource.animation.dude.thrust.right.current(t);
		} else {
			if (Math.ceil(model.moving) ) {
				if (Math.ceil(model.direction)<1)
					dudeFrame = resource.animation.dude.move.left.current(t);
				else
					dudeFrame = resource.animation.dude.move.right.current(t);
				state.model.currentFrame = dudeFrame;
			} else {
				var frame = dudeFrame % 10;
				if (Math.ceil(model.direction)<1)
					dudeFrame = frame+10;
				else
					dudeFrame = frame;
			}
		}

		//	-	-	-

		// Draw background & foreground level
		drawBackground(mapPosition.x, mapPosition.y);
		state.cuddlyLevel.render(mapPosition.x, mapPosition.y);

		// Render the red dude to the game canvas
		drawDude(dudeScreenPosition.x, dudeScreenPosition.y-bounce, dudeFrame);

		// Draw Carebears sine sprites
		sineSprites.go();

		// Draw scrolltext
		var scrollerLength = state.scrollerLevel.widthInPixels;
		state.scrollerLevel.render(Math.floor(model.scrollerPosition % scrollerLength), 0);

		//	-	-	-

		// Render now completed game canvas frame over 'framebuffer' canvas
		mycanvas.contex.drawImage(resource.ctx.game.canvas, 64, 54+32);
	}

	function drawDude(posX, posY, animFrame) {
		var a = resource.tile.dude.position(animFrame);

		resource.ctx.game.drawImage(
			resource.image.dude,
			a.x, a.y, a.w, a.h,
			Math.floor(posX), Math.floor(posY), a.w, a.h
		);
	}

	function fillBackground() {
		var s = resource.background.tile.position(1);

		for(var y=0; y<cuddlyMap.length + 1; y++)
			for(var x=0; x<cuddlyMap[0].length; x++) 
				resource.ctx.background.drawImage(resource.image.tiles, s.x, s.y, s.w, s.h, x*16*2, y*16*2, s.w, s.h);
	}

	function drawBackground(posX, posY) {
		if (posX<0) posX=0;
		if (posY<0) posY=0;

		var _SCALE = {
			X: .5,
			Y: .5 
		}

		var _SKEW = {
			X: 0,
			Y: 0
		}

		var delta = {
			x: ( _SKEW.X + (posX * _SCALE.X) ) % (16*2),
			y: ( _SKEW.Y + (posY * _SCALE.Y) ) % (16*2)
		}

		var _w = resource.ctx.background.canvas.width - 16*2;
		var _h = resource.ctx.background.canvas.height - 16*2;
		resource.ctx.game.drawImage(
			resource.ctx.background.canvas,
			Math.floor(delta.x), Math.floor(delta.y), _w, _h,
			0, 0, _w, _h 
		);

	}


	// menu API
	return {
		triggerKeyboardEvent: function(event) {
			state.keyboard.trigger(event);
		},

		bootstrap: bootstrap,
		go: runMenu
	}

})();
