'use strict'

/*
http://frutbunn.com  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

var cuddlyDemosShared = (function() {

	var globalBasePath="cuddlydemos/";


	var state = {
		toggle: {
			crtEmulator: false,
			noAspectRatio: false,
			monitor: false
		},

		autoPilot: {
			duration: null,

			setInterval: {
				loadMenu: null
			}		
		},

		canvas: null
	}

	function initAutoPilot(autoPilotDuration) {
		autoPilotDuration = autoPilotDuration || 1000*60*3;

		state.autoPilot.setInterval.loadMenu = null;
		var wasActiveLastTime = false;

		try {
			var cookieData = fruCookie.get("cuddlyMenu-autoPilot");
			if (cookieData != null) {
				wasActiveLastTime = cookieData.active;
			}
		} catch(e) {}

		if (wasActiveLastTime) {
			console.log("SCREEN IS IN AUTO PILOT MODE");
			state.autoPilot.setInterval.loadMenu = setInterval(function(){
				window.location.href = 'cuddly_menu.html';
			}, autoPilotDuration);
		} else {
			console.log("SCREEN IS ***NOT*** IN AUTO PILOT MODE");
		}

	}

	function disableAutoPilot() {
		if (state.autoPilot.setInterval.loadMenu==null)
			return;

		clearInterval(state.autoPilot.setInterval.loadMenu);
		state.autoPilot.setInterval.loadMenu=null;
		setAutoPilotCookie();
	}

	function getAutoPilotCookieNextScreen() {
		var nextScreen = 0;

		try {
			var cookieData = fruCookie.get("cuddlyMenu-autoPilot");
			if (cookieData != null) {
				nextScreen = cookieData.nextScreen;
			}
		} catch(e) {}

		return nextScreen;
	}

	function setAutoPilotCookie() {
		try {
			fruCookie.set("cuddlyMenu-autoPilot", {
				active: false,
				nextScreen: getAutoPilotCookieNextScreen()
			}, "/");
		} catch(e) {}
	}

	function initCrtEmulator(canvas, id, autoPilotDuration) {
		var gamma = 1;
		var contrast = 1.05;
		var saturation = 1.05;
		var brightness = 1;
		var light = 4;

		state.canvas = canvas;

		initAutoPilot(autoPilotDuration);

		// Get crt emulator settings
		getCookie();

		// Init crt emulator
        crtEmulator.init(canvas, id);
        crtEmulator.resizeScanlinesOffThreshold = 1;

	    crtEmulator.useCrtEmulator = state.toggle.crtEmulator;
	    crtEmulator.horizontalAspectRatio = !state.toggle.noAspectRatio;

	    crtEmulator.resizeScanlinesOffThreshold = .5;

        crtEmulator.gamma = gamma;
        crtEmulator.contrast = contrast;
        crtEmulator.saturation = saturation;
        crtEmulator.brightness = brightness;
        crtEmulator.light = light;
	}

	function getCookie() {
		try {
			var cookieData = fruCookie.get("cuddlyDemos");
			if (cookieData != null) {
				state.toggle.crtEmulator = cookieData.crtEmulator || false;
				state.toggle.noAspectRatio = cookieData.noAspectRatio || false;
				state.toggle.monitor = cookieData.monitor || false;
			}
		} catch(e) {}

	}

	// shared API
	return {
		initCrtEmulator: initCrtEmulator,
		disableAutoPilot: disableAutoPilot,

		get canvasWidth() {if (state.canvas) return state.canvas.width; else return null; },
		get canvasHeight() {if (state.canvas) return state.canvas.height; else return null; },

		path: {
			get globalBase() {return globalBasePath; }
		},

	}

})();