'use strict'

/*
http://frutbunn.com  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

var cuddlyDemosWrapper = (function() {

	// crtEmulator in iframe
	function crtEmulator() {
		var x = document.getElementById('js-iframe-cuddly-content').contentWindow.crtEmulator;
		return x;
	}

	var globalBasePath="cuddlydemos/";

	var state = {
		toggle: {
			crtEmulator: false,
			monitor: false,
			noAspectRatio: false,
			smallSize: true,
			zoom: false,
			loader: false
		}//,

		//x: false

	}

	var monitorImageSize = {
		width: 1104,
		height: 789
	}

	function init() {

		setAutoPilotCookie();

		getCookie();
		updateButtons();

		setTimeout(function() {
			$('.button').fadeIn(750);
		}, 250);

		$(window).resize(function() {
			updateScreenPosition();
		});
		updateScreenPosition();

		window.addEventListener('keydown', keyPressEvent, false);
		window.addEventListener('keyup', keyPressEvent, false);

		resizeMonitor();
		if (state.toggle.monitor)
			$('#layer-1').show();

        window.addEventListener("resize", windowResized);
        addMulitpleEventListeners(document, ['fullscreenchange', 'MSFullscreenChange', 'webkitfullscreenchange', 'mozfullscreenchange', 'ofullscreenchange'], fullscreenTriggered);
	}

	// Pass all keypress events to the iframe
	function keyPressEvent(event) {
		event.preventDefault();
		try {		
			document.getElementById('js-iframe-cuddly-content').contentWindow.cuddlyMenuScreen.triggerKeyboardEvent(event);
		} catch(e) {}

		try {
			document.getElementById('js-iframe-cuddly-content').contentWindow.KeyCheck(event);
		} catch(e) {}
	}


//	-	-	-	-	-	-	-	-	-


	function windowResized(e) {
		resizeMonitor();
	}

	function fullscreenTriggered(e) {
		if (getVendorSpecificElement(document, ['fullscreenElement', 'msFullscreenElement', 'webkitFullscreenElement', 'mozFullScreenElement', 'oFullscreenElement']) ) {
            // fullscreen
            $('#container').height("100%");
		} else {
            $('#container').height("95%");
		}

        resizeMonitor();
	}

	function resizeMonitor() {
		if (!state.toggle.monitor)
			return;

        var availableWidth = $("#container").width();
        var availableHeight = $("#container").height();
        var scale = monitorImageSize.height / monitorImageSize.width;

        // Calculate correct aspect ratio:
        var newWidth = availableWidth;
        var newHeight = Math.floor(newWidth*scale);
        var top = 0, left = 0;

        if (newHeight > availableHeight) {
            newWidth = Math.ceil( (1 / scale) * availableHeight);
            newHeight = availableHeight;
        }

        if (newWidth < availableWidth) {
            left = Math.ceil( (availableWidth-newWidth) / 2);
        }

        if (newHeight < availableHeight) {
            top = Math.ceil( (availableHeight-newHeight) / 2);
        }

        var elm = document.getElementById("layer-1");

        elm.style.position = "fixed";
       	elm.style.left = left + "px";
        elm.style.top = (top+1) + "px";
        elm.style.width = newWidth + "px";
        elm.style.height = (newHeight-1) + "px";

		var crtScale=.742;
		var crtScreenWidth = newWidth*crtScale;
		var crtScreenLeft = left + (newWidth-newWidth*crtScale)/2;
		$('#layer-2').width(crtScreenWidth);
		$('#layer-2').css("padding-left", crtScreenLeft + "px");
	}

//	-	-	-	-	-	-	-	-	-

    function addMulitpleEventListeners(node, events, func) {
        for (var i in events)
            node.addEventListener(events[i], func);
    }

    function getVendorSpecificElement(node, elements) {
        for(var i in elements)
            if (typeof node[elements[i]] !== 'undefined')
                return node[elements[i]];

        return undefined;
    }

//	-	-	-	-	-	-	-	-	-

	function updateScreenPosition() {
		var windowWidth = $(window).width();

		if (state.toggle.smallSize && windowWidth>768 ) {
			$('#layer-2').width(768);
			var paddingLeft = (windowWidth-768)/2;
			$('#layer-2').css("padding-left", paddingLeft + "px");
		} else {
			$('#layer-2').width("100%");
			$('#layer-2').css("padding-left", "0px");
		}

	}
	
	function loadScreen(screenUrl) {
		setTimeout(function(){
			document.getElementById("js-iframe-cuddly-content").src = screenUrl;
			//document.getElementById("js-iframe-cuddly-content").reload(true);

		}, 250);

	}

	function getAutoPilotCookieNextScreen() {
		var nextScreen = 0;

		try {
			var cookieData = fruCookie.get("cuddlyMenu-autoPilot");
			if (cookieData != null) {
				nextScreen = cookieData.nextScreen;
			}
		} catch(e) {}

		return nextScreen;
	}

	function setAutoPilotCookie() {
		try {
			fruCookie.set("cuddlyMenu-autoPilot", {
				active: false,
				nextScreen: getAutoPilotCookieNextScreen()
			}, "/");
		} catch(e) {}
	}

	function launchReset() {
		setAutoPilotCookie();
		loadScreen('cuddly_reset.html');	
	}

	function launchMenu() {
		setAutoPilotCookie();
		loadScreen('cuddly_menu.html');
	}

	function getCookie() {
		try {
			var cookieData = fruCookie.get("cuddlyDemos");
			if (cookieData != null) {
				state.toggle.crtEmulator = cookieData.crtEmulator;
				state.toggle.monitor = cookieData.monitor;
				state.toggle.noAspectRatio = cookieData.noAspectRatio;
				state.toggle.smallSize = cookieData.smallSize;
				state.toggle.zoom = cookieData.zoom;
			}
		} catch(e) {}

	}

	function setCookie() {
		try {
			fruCookie.set("cuddlyDemos", {
				crtEmulator: state.toggle.crtEmulator,
				monitor: state.toggle.monitor,
				noAspectRatio: state.toggle.noAspectRatio,
				smallSize: state.toggle.smallSize,
				zoom: state.toggle.zoom
			}, "/", 20*365);
		} catch(e) {}

	}

	function updateButtons() {
		var buttons = {
			"#cuddly-demos-crtEmulator": {
				disabled: false,
				selected: false
			},
			"#cuddly-demos-monitor": {
				disabled: false,
				selected: false
			},
			"#cuddly-demos-no-aspect-ratio": {
				disabled: false,
				selected: false
			},
			"#cuddly-demos-small-size": {
				disabled: false,
				selected: false
			},
			"#cuddly-demos-zoom": {
				disabled: false,
				selected: false
			}
		
		}

		// Handle buttons:
		buttons["#cuddly-demos-crtEmulator"].selected = state.toggle.crtEmulator;
		buttons["#cuddly-demos-monitor"].selected = state.toggle.monitor;
		buttons["#cuddly-demos-no-aspect-ratio"].selected = state.toggle.noAspectRatio;
		buttons["#cuddly-demos-small-size"].selected = state.toggle.smallSize;
		buttons["#cuddly-demos-zoom"].selected = state.toggle.zoom;

		// Update buttons
		for(var name in buttons) {
			// All buttons are unselected, and enabled
			$(name).removeClass('selected selected2').removeClass('disabled').attr("disabled", false);

			// Disable button
			if (buttons[name].disabled)
				$(name).addClass('disabled').attr("disabled", true);

			// Set button as selected, but only if enabled
			if (buttons[name].selected) {
				if (name=="#cuddly-demos-crtEmulator") 
					$(name).addClass('selected2');
				else
					$(name).addClass('selected');
			}
		}

		setCookie();
		updateScreenPosition();
		resizeMonitor();
		giveIframeFocus();
	}

    function giveIframeFocus() {
        var elm = $("#js-iframe-cuddly-content")[0];
        elm.contentWindow.focus();
    }

	function toggleCrtEmulator() {
	    state.toggle.crtEmulator = !state.toggle.crtEmulator;
	    crtEmulator().useCrtEmulator = state.toggle.crtEmulator;
	    updateButtons();
	}

	//	-	-	-
	// Please ignore the mess below - I only had 10 minutes to convert it to a radio option...
	//	-	-	-

	function toggleMonitor() {
		state.toggle.zoom = false;
	    state.toggle.monitor = true;
	    state.toggle.smallSize = false;
	    state.toggle.noAspectRatio = false;

	    crtEmulator().horizontalAspectRatio = false;
	    $('#layer-1').show();
	    updateButtons();
	}

	function noAspectRatio() {
		$('#layer-1').hide();

		state.toggle.zoom = false;
		state.toggle.monitor = false;
		state.toggle.smallSize = false;
		state.toggle.noAspectRatio = true

		crtEmulator().horizontalAspectRatio = false;
		updateButtons();	
	}

	function smallSize() {
		$('#layer-1').hide();

		state.toggle.zoom = false;
		state.toggle.monitor = false;
		state.toggle.smallSize = true;
		state.toggle.noAspectRatio = false;

		crtEmulator().horizontalAspectRatio = false;
		updateButtons();
	}

	function zoom() {
		$('#layer-1').hide();

		state.toggle.zoom = true;
		state.toggle.monitor = false;
		state.toggle.smallSize = false;
		state.toggle.noAspectRatio = false;

		crtEmulator().horizontalAspectRatio = true;
		updateButtons();
	}

	function loader(that) {
		state.toggle.loader = !state.toggle.loader;
		$(that).toggleClass("selected");

		if (state.toggle.loader)
			$('#layer-3').fadeIn(400);
		else
			$('#layer-3').fadeOut(400, function() {
				$('#layer-3').hide();
			});

	}

	function run(that, demoUrl) {
		setAutoPilotCookie();

		state.toggle.loader = false;
		$("#js-loader-button").removeClass("selected");

		$(that).addClass("have-ran");

		$('#layer-3').fadeOut(400, function() {
			$('#layer-3').hide();

			loadScreen(demoUrl);
		});

	}

	//	-	-	-

	// shared API
	return {
		init: init,

		buttonClick: {
			toggle: {
				crtEmulator: toggleCrtEmulator,
				monitor: toggleMonitor,
				noAspectRatio: noAspectRatio,
				smallSize: smallSize,
				zoom: zoom,
				loader: loader,
				run: run
			},
			launch: {
				resetScreen: launchReset,
				menuScreen: launchMenu	
			}
		},

	}

})();