'use strict'

/*
http://frutbunn.tk  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

// frameTimeThreshold & userObject are optional.
function FruTimeStepping(deltaTime, stateObject, integrateCallback, renderCallback, userObject, frameTimeThreshold) {
	var that = this;

	var DELTA_TIME_MULTIPLE = 5;

	var state = {
		t: 0,
		dt: undefined,

		currentTime: null,
		accumulator: 0,
		frameTimeThreshold: undefined,

		userObject: undefined
	}

	var model = {
		current: undefined,
		previous: undefined
	}

	var callback = {
		integrate: undefined,
		render: undefined
	}

	//	-	-	-

	function constructor() {
		state.dt = deltaTime;

		model.current = stateObject;
		model.previous = cloneObject(model.current);

		callback.integrate = integrateCallback;
		callback.render = renderCallback;

		state.frameTimeThreshold = frameTimeThreshold || deltaTime*DELTA_TIME_MULTIPLE;

		state.userObject = userObject;
	}

	//	-	-	-

	function isObject(x) { return x && typeof x === 'object' && x.constructor !== Array; }
	function isArray(x) { return x && typeof x === 'object' && x.constructor === Array; }
	function isPrimative(x) { return ( !x || (x && typeof x !== 'object') ); }

	function tree(name) {return name ? name + "." : "";}

	function cloneObject(source) {
		var obj = blendObject(source, source, 1);
		return obj;
	}

	function blendObject(current, previous, alpha) {
//		console.log("->blendObject");
		function _blend(dest, current, previous, name) {
//			console.log("->_blend()");
			for (var hash in current) {
//				console.log(tree(name) + hash);
				if (isObject(current[hash])) {
					dest[hash] = {};
					_blend(dest[hash], current[hash], previous[hash], tree(name) + hash);
				} else if (isArray(current[hash])) {
					var d = dest[hash] = [];
					var c = current[hash];
					var p = previous[hash];

					for(var i in c) {
						var n = c[i];

						if (isObject(n) ) {
							d[i] = {};
							_blend(d[i], c[i], p[i], tree(i) + hash);
						} else {
							d[i] = c[i]*alpha + p[i]*(1-alpha);
						}

					}

				} else if (isPrimative(current[hash])) {
					dest[hash] = current[hash]*alpha + previous[hash]*(1-alpha);
				} else {
					throw new Error("FixedTimeStep:blendObject: Property is not an object, array or primative - bug...");
				}

			}
//			console.log("<-_blend()");

		}

		var obj = {};
		_blend(obj, current, previous);

		return obj;
	}

	function go(newTime) {
		if (state.currentTime === null)
			state.currentTime = newTime-state.dt;

		var frameTime = newTime - state.currentTime;
		if (frameTime > state.frameTimeThreshold)
			frameTime = state.frameTimeThreshold;
		state.currentTime = newTime;

		state.accumulator += frameTime;

		while(state.accumulator >= state.dt) {
			model.previous = cloneObject(model.current);
			integrate();

			state.t += state.dt;
			state.accumulator -= state.dt;
		}

		var alpha = state.accumulator/state.dt;
		var blendedModel = blendObject(model.current, model.previous, alpha);
		render(blendedModel);
	}

	function integrate() {
		try { callback.integrate(model.current, state.t, state.dt, state.userObject); }
		catch(e) { throw new Error("FixedTimeStep: Could not call integrate callback.\n" + e); }
	}

	function render(blendedModel) {
		try { callback.render(blendedModel, state.t, state.dt, state.userObject); }
		catch(e) { throw new Error("FixedTimeStep: Could not call render callback.\n" + e); }
	}

	//	-	-	-

	constructor();

	return {
		go: go
	}

}