'use strict'

/*
http://frutbunn.tk  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

var cuddlyMenuScreen = (function() {

	var player = new music("YM");

	var state = {
		keyboard: undefined,
		globalTime: new FruTimer(),
		carebaresGlobalTime: new FruTimer(),

		stepping: undefined,

		cuddlyLevel: undefined,
		scrollerLevel: undefined,

		// Everything in model will be subject to linear interpolation by fru_time_stepping.js
		model: {
			position: {	// Position of dude on the scrolling map
				x: 320,
				y: 450
			},

			direction: 1, // Direction red dude is facing 0 = left, 1 = right
			moving: 0,
			thrusting: 0,

			justLanded: 0,
			bounceDisplacement: 0,

			thrustSpeed: 0,
			fallingSpeed: 0,

			currentFrame: 6,

			carebears: {
				x: [0,0,0,0,0,0,0,0,0,0,0,0],
				y: [0,0,0,0,0,0,0,0,0,0,0,0]
			},

			scrollerPosition: 0
		},


	};

	var resource = {
		totalImagesToLoad: 4,

		image: undefined,	// image is an object setup in bootstrap(), and looks like:
//		image: {
//			carebears: 'carebears.png' image,
//			dude: 'dude.png' image,
//			tiles: 'tiles.png' image,
//			chrome: 'chrome.png' image
//		}

		ctx: {
			framebuffer: undefined,
			game: undefined,
			background: undefined
		},

		tile: {
			dude: new FruTools.Tile(640, 64, 64),
			carebears: new FruTools.Tile(384, 32, 20)
		},

		animation: {
			dude: {
				move: {
					right: new FruTools.Animation(.35, [2, 3, 4, 5, 6, 7, 8, 9], true),
					left: new FruTools.Animation(.35, [12, 13, 14, 15, 16, 17, 18, 19], true)
				},
				thrust: {
					right: new FruTools.Animation(.075, [0, 1], true),
					left: new FruTools.Animation(.075, [10, 11], true)
				}
			}
		},

		background: {
			tile: new FruTools.Tile(320, 16*2, 16*2, 10)
		},

		bouncing: {
				animation: [0,3,5,6,5,3,0,1,2,3,2,1,0],
				speed: 7
		}

	}

	function anImageLoaded() {
	    if (--resource.totalImagesToLoad>0)
	        console.log("An image loaded. (Still waiting for " + resource.totalImagesToLoad + " resources to load).");
	    else
			console.log("All resources loaded...");
	}

	function loadImage(url) {
		console.log("Loading image: " + url);
		var img = document.createElement('img');
		img.src = url;
		img.addEventListener('load', anImageLoaded);

		return img;
	}

	function createCanvasElement(width, height) {
		var canvas = document.createElement('canvas');
		canvas.width  = width;
		canvas.height = height;
		return canvas.getContext("2d");
	}

	var requestAnimFrame = (function() {
	   	return window.requestAnimationFrame || window.webkitRequestAnimationFrame || window.mozRequestAnimationFrame 
	   	 || window.oRequestAnimationFrame || window.msRequestAnimationFrame
	   	 || function(callback, element) { window.setTimeout(callback, 1000/60); };
	})();

	// The first function that runs - see index.html body tag
	function bootstrap() {
		resource.ctx.framebuffer = createCanvasElement(768, 540);

		var elm = document.getElementById('menu');
		elm.appendChild(resource.ctx.framebuffer.canvas);

		resource.ctx.game = createCanvasElement(320*2, 144*2);
		resource.ctx.background = createCanvasElement(320*2 + 16*2, cuddlyMap.length*16*2 + 16*2); // Background width + 1 tile

		//	-	-	-

		// Do not move from bootstrap(). Otherwise the onload event in the HTML will not
		// be triggered until these images have loaded.
		resource.image = {
			carebears: loadImage(basepath + 'carebears.png'),
			dude: loadImage(basepath + 'dude.png'),
			tiles: loadImage(basepath + 'tiles.png'),
			chrome: loadImage(basepath + 'chrome.png')
		}

		//	-	-	-

		// Initialize Carl's loader screen: 
		loaderInit();

		//	-	-	-

		// Show a black screen until all images have been loaded
		goLoadingImages();
	}

	// Black screen while loading image resources
	function goLoadingImages() {
		if (state.globalTime.seconds.current>.1 && resource.totalImagesToLoad<=0) {
			main();

			return;
		}

		//	-	-	-

		var w = resource.ctx.framebuffer.canvas.width;
		var h = resource.ctx.framebuffer.canvas.height;
		resource.ctx.framebuffer.fillStyle="#000";
		resource.ctx.framebuffer.fillRect(0, 0, w, h);

		//	-	-	-

		requestAnimFrame(goLoadingImages);
	}

	function sleep(duration, callback) {
		setTimeout(callback, duration);
	}

	function main() {
		state.keyboard = new FruKeyboard({
			"37": "left",
			"90": "left",
			"39": "right",
			"88": "right",
			"38": "thrust",
			"13": "thrust",
			"32": "load",


			"98": "down",
			"104": "up"
		});

		state.stepping = new FruTimeStepping(1/60, state.model, integrate, render, state);
		state.cuddlyLevel = new FruRenderMap(resource.image.tiles, cuddlyMap, 16*2
		 , resource.ctx.game.canvas,    0, 0, 320*2, 144*2);
		
		var scrollerMap = buildScrolltextData(scrolltextData); // scrolltextData is in data.js
		state.scrollerLevel = new FruRenderMap(resource.image.chrome, scrollerMap
		 , [16*2, 40*2], resource.ctx.framebuffer.canvas, 64, 400, 640, 80);
		
		// Precalculate background scrolling area
		fillBackground();

		// See if coming back to a previous position on the map / scrolltext
		try {
			var cookieData = fruCookie.get("cuddlyMenu");
			if (cookieData != null) {
				state.model.position.x = cookieData.locationX;
				state.model.position.y = cookieData.locationY-50;
				state.model.scrollerPosition = cookieData.scrollerPosition;
			}
		} catch(e) {}

		//	-	-	-

		state.globalTime.reset();
		state.carebaresGlobalTime.reset();

		sineSprites.init(state.carebaresGlobalTime, resource.image.carebears, resource.tile.carebears, resource.ctx.game);

		// -	-	-

		// Run Carl's loader screen in loader.js
		loader("MENU");
	}

	// Carl's code will call this function to run the menu.
	function runMenu() {

//		sleep(750, function(){
			player.LoadAndRun(basepath + 'menu.ym');
			document.getElementById("menu").style.display = "block";
			go();
//		});

	}

	function buildScrolltextData(text) {
		var scrollerCharWidth = [
		3, 1, 2, 3, 3, 3, 3, 1, 1, 1, 3, 2, 1, 2, 1, 2, 3, 2, 2, 2, 2, 2, 
		2, 2, 2, 2, 1, 1, 3, 2, 3, 2, 3, 
		3, 2, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 3, 2, 3, 2, 3, 2, 2, 2, 2, 3, 3, 3, 2, 2,
		3, 3, 3, 3, 3, 3
		];

		var scroller = [];
		for(var i=0; i<text.length; i++) {
			var char = text.charCodeAt(i);
			var p = (char-32);
			var blocks = scrollerCharWidth[p];

			for(var j=0; j<blocks; j++)
				scroller.push(p*3 + j);
		}

		var map = [];
		map.push(scroller);
		return map;
	}

	//	-	-	-

	function DemoScreenLocation(x, y, screen) {
		this.x = x;
		this.y = y;
		this.screen = screen;
	}

	var demoScreens = [
		new DemoScreenLocation(27, 7, "BIG_SPRITE"),
		new DemoScreenLocation(63, 16, "COLORSHOCK_II"),
		new DemoScreenLocation(87, 6, "NO_NAME_1"),
		new DemoScreenLocation(128, 1, "MEGA_SCROLLER"),
		new DemoScreenLocation(144, 8, "SPREADPOINT"),
		new DemoScreenLocation(168, 16, "DIGI_DEMO"),
		new DemoScreenLocation(197, 2, "LED_SCROLLER"),
		new DemoScreenLocation(294, 2, "DOC"),
		new DemoScreenLocation(351, 13, "FULLSCREEN"),
		new DemoScreenLocation(373, 5, "STARWARS_DEMO"),
		new DemoScreenLocation(376, 16, "KNUCKLE_BUSTER"),
		new DemoScreenLocation(416, 14, "DNA_DEMO"),
		new DemoScreenLocation(438, 1, "NO_NAME_2")
	];

	function loadScreen() {
		var load = state.keyboard.isDown("load");
		if (load) {

			var p = new FruTools.Coordinate(
			 Math.floor(state.model.position.x/32), Math.floor(state.model.position.y/32)
			);

			for(var i in demoScreens) {
				var t = demoScreens[i];

				if (p.x>=t.x && p.x<=t.x+3 && p.y>=t.y && p.y<t.y+3) {

					// Store position of dude in a cookie so we know where we were when we come back
					try {
						fruCookie.set("cuddlyMenu", {
							locationX: state.model.position.x,
							locationY: state.model.position.y,
							scrollerPosition: state.model.scrollerPosition
						}, "/");
					} catch(e) {}

					// Stop music
					player.LoadAndRun(basepath + 'mute.ym');

					// Clear screen:
					var w = resource.ctx.framebuffer.canvas.width;
					var h = resource.ctx.framebuffer.canvas.height;
					resource.ctx.framebuffer.fillStyle="#000";
					resource.ctx.framebuffer.fillRect(0, 0, w, h);					

					// Run Carl's loader screen:
					sleep(500, function(){
						// Hide my screen:
						document.getElementById("menu").style.display = "none";
						loader(t.screen);
					});

					return false;
				}

			}

		}

		return true;
	}

	//	-	-	-

	// Main game loop:
	function go() {
		var time = state.globalTime.seconds.current;
		state.stepping.go(time);

		if (loadScreen() ) {
			requestAnimFrame(go);
		}
	}

	//	-	-	-

	function integrate(model, t, dt, userObject) {
		var left = state.keyboard.isDown("left");
		var right = state.keyboard.isDown("right");
		var thrust = state.keyboard.isDown("thrust");

		if (model.bounceDisplacement>0)
			model.bounceDisplacement-=1;

		if (left && !right) {
			model.position.x -= resource.bouncing.speed;
			model.direction = 0;
			model.moving = 1;
		}

		if (right && !left) {
			model.position.x += resource.bouncing.speed;
			model.direction = 1;
			model.moving = 1;
		}

		if (!right && !left) {
			model.moving = 0;
		}

		if (thrust) {
			model.thrusting = 1;

			if (model.thrustSpeed<=0)
				model.thrustSpeed = 3;

			if (model.thrustSpeed<resource.bouncing.speed )
				model.thrustSpeed += 1;

			model.fallingSpeed = 0;
		} else {
			model.thrusting = 0;

			if (model.position.y<32)
				model.thrustSpeed = 0;
		}


		if (model.thrustSpeed<=0 && model.justLanded!=1) { 
			if (model.fallingSpeed<resource.bouncing.speed)
				model.fallingSpeed +=  .5;
		}


		for(var i=0; i< Math.floor(model.fallingSpeed)+1; i++) {
			if (! haveLanded(model) ) {
				model.position.y += 1; 
				model.justLanded = 0;
			} else {
				if (model.justLanded < 1) {
					model.justLanded = 1;

					model.bounceDisplacement = resource.bouncing.animation.length;
					model.thrustSpeed = 0;
				}

			}

		}

		if (model.thrustSpeed>0) 
			model.thrustSpeed -= .5;

		if (model.thrustSpeed>0)
			model.position.y -= Math.floor(model.thrustSpeed);

		// clipping
		var DUDE_WIDTH = resource.tile.dude.width;
		var DUDE_HEIGHT = resource.tile.dude.height;
		var MAP_WIDTH_IN_PIXELS = state.cuddlyLevel.widthInPixels;
		var MAP_HEIGHT_IN_PIXELS = state.cuddlyLevel.heightInPixels - (32-4);

		// Same as Cuddly Menu
		if (model.position.x<9*32) 
			model.position.x = 9*32;
		
		if (model.position.y<0) 
			model.position.y = 0;
		
		if (model.position.x>=455*32) 
			model.position.x = 455*32;
		
		if (model.position.y>=MAP_HEIGHT_IN_PIXELS-DUDE_HEIGHT) 
			model.position.y = MAP_HEIGHT_IN_PIXELS-DUDE_HEIGHT;



		integrateScroller(model, t, dt, userObject);
	}

	function integrateScroller(model, t, dt, userObject) {
		model.scrollerPosition += 8;
	}

	function haveLanded(model) {
		var FLOOR_TILE = 69;

		var dudePosition = new FruTools.Coordinate(model.position.x, model.position.y);
		var topLeftBlock = new FruTools.Coordinate( Math.floor(dudePosition.x/32), Math.floor(dudePosition.y/32) );
		
		var collision = false;

		if ( 
		 (cuddlyMap[topLeftBlock.y+2][topLeftBlock.x]==FLOOR_TILE
		 || cuddlyMap[topLeftBlock.y+2][topLeftBlock.x+1]==FLOOR_TILE
		 || (topLeftBlock.x<cuddlyMap[0].length-2 && cuddlyMap[topLeftBlock.y+2][topLeftBlock.x+2]==FLOOR_TILE)
		 ) 
		 && model.position.y%32 == 0
		) collision = true;

		return collision
	}

	function render(model, t, dt, userObject) {
		var w = resource.ctx.framebuffer.canvas.width;
		var h = resource.ctx.framebuffer.canvas.height;
		resource.ctx.framebuffer.fillStyle="#000";
		resource.ctx.framebuffer.fillRect(0, 0, w, h);

		var DUDE_WIDTH = resource.tile.dude.width;
		var DUDE_HEIGHT = resource.tile.dude.height;
		var MAP_WIDTH_IN_PIXELS = state.cuddlyLevel.widthInPixels;
		var MAP_HEIGHT_IN_PIXELS = state.cuddlyLevel.heightInPixels - (32-4);

		var canvas = resource.ctx.game.canvas;

		var dudePosition = new FruTools.Coordinate(Math.round(model.position.x), Math.round(model.position.y) );
		var dudeScreenPosition = new FruTools.Coordinate();
		var mapPosition = new FruTools.Coordinate();

		// Calculate position of dude, map and background:
		if (dudePosition.x <= canvas.width/2 - DUDE_WIDTH/2) {
			dudeScreenPosition.x = dudePosition.x;
			mapPosition.x = 0;
		} else if (dudePosition.x > MAP_WIDTH_IN_PIXELS - canvas.width/2 - DUDE_WIDTH/2) {
			dudeScreenPosition.x = canvas.width - (MAP_WIDTH_IN_PIXELS-dudePosition.x);
			mapPosition.x = MAP_WIDTH_IN_PIXELS - canvas.width;
		} else {
			dudeScreenPosition.x = canvas.width/2 - DUDE_WIDTH/2;
			mapPosition.x = dudePosition.x - (canvas.width/2 - DUDE_WIDTH/2);
		}

		if (dudePosition.y <= canvas.height/2 - DUDE_HEIGHT/2) {
			dudeScreenPosition.y = dudePosition.y;
			mapPosition.y = 0;
		} else if (dudePosition.y > MAP_HEIGHT_IN_PIXELS - canvas.height/2 - DUDE_HEIGHT/2) {
			dudeScreenPosition.y = canvas.height - (MAP_HEIGHT_IN_PIXELS-dudePosition.y);
			mapPosition.y = MAP_HEIGHT_IN_PIXELS - canvas.height;
		} else {
			dudeScreenPosition.y = canvas.height/2 - DUDE_HEIGHT/2;
			mapPosition.y = dudePosition.y - (canvas.height/2 - DUDE_HEIGHT/2);
		}

		// Calculate dude's landing bounce:
		var bounce=resource.bouncing.animation[resource.bouncing.animation.length-Math.floor(model.bounceDisplacement)-1];
		bounce *= ( (model.fallingSpeed<3 ? 3 : model.fallingSpeed) / resource.bouncing.speed) * .9;

		// Calculate red dude frame of animation
		var dudeFrame = state.model.currentFrame;
		if (Math.ceil(model.thrusting) ) {
			if (Math.ceil(model.direction)<1)
				dudeFrame = resource.animation.dude.thrust.left.current(t);
			else
				dudeFrame = resource.animation.dude.thrust.right.current(t);
		} else {
			if (Math.ceil(model.moving) ) {
				if (Math.ceil(model.direction)<1)
					dudeFrame = resource.animation.dude.move.left.current(t);
				else
					dudeFrame = resource.animation.dude.move.right.current(t);
				state.model.currentFrame = dudeFrame;
			} else {
				var frame = dudeFrame % 10;
				if (Math.ceil(model.direction)<1)
					dudeFrame = frame+10;
				else
					dudeFrame = frame;
			}
		}

		//	-	-	-

		// Draw background & foreground level:
		drawBackground(mapPosition.x, mapPosition.y);
		state.cuddlyLevel.render(mapPosition.x, mapPosition.y);

		// Render the red dude to the game canvas
		drawDude(dudeScreenPosition.x, dudeScreenPosition.y-bounce, dudeFrame);

		// Draw Carebears sine sprites:
		sineSprites.go();

		// Draw scrolltext:
		var scrollerLength = state.scrollerLevel.widthInPixels;
		state.scrollerLevel.render(Math.round(model.scrollerPosition % scrollerLength), 0);

		//	-	-	-

		// Render now completed game canvas frame over 'framebuffer' canvas:
		resource.ctx.framebuffer.drawImage(resource.ctx.game.canvas, 64, 54+32);
	}

	function drawDude(posX, posY, animFrame) {
		var a = resource.tile.dude.position(animFrame);

		resource.ctx.game.drawImage(
			resource.image.dude,
			a.x, a.y, a.w, a.h,
			posX, posY, a.w, a.h
		);
	}

	function fillBackground() {
		var s = resource.background.tile.position(1);

		for(var y=0; y<cuddlyMap.length + 1; y++)
			for(var x=0; x<cuddlyMap[0].length; x++) 
				resource.ctx.background.drawImage(resource.image.tiles, s.x, s.y, s.w, s.h, x*16*2, y*16*2, s.w, s.h);
	}

	function drawBackground(posX, posY) {
		if (posX<0) posX=0;
		if (posY<0) posY=0;

		var _SCALE = {
			X: .5,
			Y: .5 
		}

		var _SKEW = {
			X: 0,
			Y: 0
		}

		var delta = {
			x: ( _SKEW.X + (posX * _SCALE.X) ) % (16*2),
			y: ( _SKEW.Y + (posY * _SCALE.Y) ) % (16*2)
		}

		var _w = resource.ctx.background.canvas.width - 16*2;
		var _h = resource.ctx.background.canvas.height - 16*2;
		resource.ctx.game.drawImage(
			resource.ctx.background.canvas,
			delta.x, delta.y, _w, _h,
			0, 0, _w, _h 
		);

	}

	// menu API: 
	return {
		bootstrap: bootstrap,
		go: runMenu
	}

})();
